"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import TimerParams

from .when import HomieWhen

if TYPE_CHECKING:
    from .homie import Homie
    from .params import HomieDesireParams
    from .params import HomieWhenParams



_OUTCOMES = dict[str, list[bool]]



class HomieDesire:
    """
    Normalize the desired parameter across multiple products.

    :param name: Name of the object within the Homie config.
    :param homie: Primary class instance for Homie Automate.
    """

    __homie: 'Homie'
    __params: 'HomieDesireParams'

    __name: str

    __when: list[HomieWhen]


    def __init__(
        self,
        homie: 'Homie',
        name: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.log_d(
            base='homie',
            item='HomieDesire',
            name=name,
            status='initial')


        desires = (
            homie.params.desires)

        assert desires is not None

        params = desires[name]


        self.__homie = homie
        self.__params = params
        self.__name = name


        when = self.when or []

        self.__when = [
            HomieWhen(homie, x)
            for x in when]


        self.__validate_params()

        homie.log_d(
            base='homie',
            item='HomieDesire',
            name=name,
            status='created')


    def __validate_params(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        groups = self.homie.groups
        _groups = self.params.groups

        for name in _groups:
            assert name in groups

        scenes = self.homie.scenes
        _scene = self.params.scene

        assert _scene in scenes


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'HomieDesireParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    @property
    def whens(
        self,
    ) -> list[HomieWhen]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return list(self.__when)


    @property
    def groups(
        self,
    ) -> list[str]:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.groups


    @property
    def scene(
        self,
    ) -> str:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.scene


    @property
    def weight(
        self,
    ) -> int:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.weight


    @property
    def delay(
        self,
    ) -> int:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.delay


    @property
    def when(
        self,
    ) -> Optional[list['HomieWhenParams']]:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.when


    @property
    def delayed(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether desire is delayed.

        :returns: Boolean indicating whether desire is delayed.
        """

        homie = self.homie
        timers = homie.timers['desire']

        children = timers.children

        unique = self.name
        delay = self.delay

        if unique not in children:

            params = TimerParams(
                timer=delay,
                start=f'-{delay}s')

            timers.create(
                unique, params)

        return not timers.ready(
            unique, update=False)


    def update_timer(
        self,
    ) -> None:
        """
        Update the existing timer from mapping within the cache.
        """

        homie = self.homie
        timers = homie.timers['desire']

        timers.update(self.name, 'now')


    @property
    def outcomes(
        self,
    ) -> Optional[_OUTCOMES]:
        """
        Return the dictionaries of boolean groups from conditons.

        :returns: Dictionaries of boolean groups from conditons.
        """

        matched: _OUTCOMES = {}


        for when in self.__when:

            outcome = when.outcome
            family = when.family

            if family not in matched:
                matched[family] = []

            (matched[family]
                .append(outcome))


        return matched or None


    @property
    def outcome(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether conditions matched.

        :returns: Boolean indicating whether conditions matched.
        """

        outcomes = self.outcomes

        if outcomes is None:
            return True

        matched: list[bool] = []


        items = outcomes.items()

        for family, when in items:

            matched.append(
                all(when)
                if family == 'default'
                else any(when))

        assert len(matched) >= 1


        return all(matched)


    def homie_dumper(
        self,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        :returns: Content related to the project dumper script.
        """

        params = (
            self.params.model_dump())

        return {
            'name': self.name,
            'params': params}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .bridge import PhueBridge
from .device import PhueDevice
from .params import PhueDeviceParams
from .params import WhenPhueChangeParams
from .when import chck_phue_change
from .when import when_phue_change



__all__ = [
    'PhueBridge',
    'PhueDevice',
    'PhueDeviceParams',
    'WhenPhueChangeParams',
    'chck_phue_change',
    'when_phue_change']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Literal
from typing import Optional

from pydantic import BaseModel



_SENSORS = Literal[
    'button1',
    'button2',
    'button3',
    'button4',
    'motion1']



class PhueDeviceParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param bridge: Name of bridge where the device belongs.
    :param phid: Unique identifier for device within bridge.
    :param label: Name of the device to find within bridge.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    bridge: str
    phid: Optional[str] = None

    label: Optional[str] = None


    def __init__(
        self,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(**data)

        phid = self.phid
        label = self.label

        assert phid or label



class WhenPhueChangeParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param devices: List of devices in scope for conditional.
    :param sensors: List of sensors in scope for conditional.
    :param since: Minimum required time post change occurred.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    devices: list[str]
    sensors: list[Literal[
        'button1',
        'button2',
        'button3',
        'button4',
        'motion1']]

    since: int = 0



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from pathlib import Path
from typing import Any

from encommon.times import Duration
from encommon.times import Times
from encommon.utils import read_text
from encommon.utils import save_text

from pytest import fixture

from requests_mock import Mocker

from .config import Config
from .homie import Homie
from .homie.test import (
    SAMPLES as CORE_SAMPLES)
from .philipshue.test import (
    SAMPLES as PHUE_SAMPLES)
from .ubiquiti.test import (
    SAMPLES as UBIQ_SAMPLES)



_ANCHOR = Times('-0s@s')

REPLACES = {


    '__STAMP_SMP__': (
        _ANCHOR.shift('-1h').simple),

    '__STAMP_SUB__': (
        _ANCHOR.shift('-1h').subsec),

    '__DURATION__': (
        str(Duration(
            Times(0).since,
            groups=2))),


    '__FSEEN_CUR__': (
        str(int(_ANCHOR.shift('-9d')))),

    '__LSEEN_CUR__': (
        str(int(_ANCHOR.shift('-1d')))),

    '__FSEEN_OLD__': (
        str(int(_ANCHOR.shift('-9d')))),

    '__LSEEN_OLD__': (
        str(int(_ANCHOR.shift('-1d')))),


    '__ASSOC_TIME__': (
        str(int(_ANCHOR.shift('-2d')))),


    '__DISCO_CUR__': (
        str(int(_ANCHOR.shift('-3h')))),

    '__DISCO_OLD__': (
        str(int(_ANCHOR.shift('-1d')))),


    '__START_DAY__': (
        _ANCHOR.shift('-0d@d').simple),

    '__STOP_DAY__': (
        _ANCHOR.shift('+1d@d').simple)}



def config_factory(
    tmp_path: Path,
) -> Config:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    save_text(
        f'{tmp_path}/config.yml',
        content=(
            "cache: ':memory:'\n"
            'enconfig:\n'
            '  paths:\n'
            f'    - {tmp_path}\n'
            f'    - {CORE_SAMPLES}\n'
            f'    - {PHUE_SAMPLES}\n'
            f'    - {UBIQ_SAMPLES}\n'
            'enlogger:\n'
            '  stdo_level: info\n'))

    return Config(
        f'{tmp_path}/config.yml')



@fixture
def config(
    tmp_path: Path,
) -> Config:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    return config_factory(tmp_path)



def homie_factory(  # noqa: CFQ001
    config: 'Config',
) -> Homie:
    """
    Construct the instance for use in the downstream tests.

    .. note::
       Function has slowly evolved and grown over time, but
       should be refactored and simplified in the future.

    :param config: Primary class instance for configuration.
    :returns: Newly constructed instance of related class.
    """

    homie = Homie(config)

    bridges = homie.phue_bridges
    routers = homie.ubiq_routers


    phue_paths = [
        ('https://192.168.1.10'
         '/clip/v2/resource'),
        ('https://192.168.2.10'
         '/clip/v2/resource')]

    phue_files = [
        f'{PHUE_SAMPLES}/jupiter.json',
        f'{PHUE_SAMPLES}/neptune.json']


    ubiq_paths = [

        ('https://192.168.1.1'
         '/api/auth/login'),
        ('https://192.168.1.1'
         '/proxy/network'
         '/api/s/default/rest/user'),
        ('https://192.168.1.1'
         '/proxy/network'
         '/api/s/default/stat/sta'),

        ('https://192.168.2.1'
         '/api/auth/login'),
        ('https://192.168.2.1'
         '/proxy/network'
         '/api/s/default/rest/user'),
        ('https://192.168.2.1'
         '/proxy/network'
         '/api/s/default/stat/sta')]

    ubiq_files = [

        (f'{UBIQ_SAMPLES}/jupiter'
         '/historic.json'),
        (f'{UBIQ_SAMPLES}/jupiter'
         '/realtime.json'),

        (f'{UBIQ_SAMPLES}/neptune'
         '/historic.json'),
        (f'{UBIQ_SAMPLES}/neptune'
         '/realtime.json')]


    def _replaces(
        path: str,
    ) -> dict[str, Any]:

        content = read_text(path)

        items = REPLACES.items()

        for key, new in items:

            content = (
                content
                .replace(key, new))

        loaded = loads(content)

        assert isinstance(loaded, dict)

        return loaded


    with Mocker() as mocker:


        dumped = _replaces(phue_files[0])

        mocker.get(
            url=phue_paths[0],
            json=dumped)


        dumped = _replaces(phue_files[1])

        mocker.get(
            url=phue_paths[1],
            json=dumped)


        for bridge in bridges.values():
            assert bridge.fetched


        mocker.post(ubiq_paths[0])


        dumped = _replaces(ubiq_files[0])

        mocker.register_uri(
            method='get',
            url=ubiq_paths[1],
            response_list=[
                {'json': dumped,
                 'status_code': 401},
                {'json': dumped}])


        dumped = _replaces(ubiq_files[1])

        mocker.get(
            url=ubiq_paths[2],
            json=dumped)


        mocker.post(ubiq_paths[3])


        dumped = _replaces(ubiq_files[2])

        mocker.register_uri(
            method='get',
            url=ubiq_paths[4],
            response_list=[
                {'json': dumped,
                 'status_code': 401},
                {'json': dumped}])


        dumped = _replaces(ubiq_files[3])

        mocker.get(
            url=ubiq_paths[5],
            json=dumped)


        for router in routers.values():
            assert router.fetched


    return homie



@fixture
def homie(
    config: 'Config',
) -> Homie:
    """
    Construct the instance for use in the downstream tests.

    :param config: Primary class instance for configuration.
    :returns: Newly constructed instance of related class.
    """

    return homie_factory(config)
